#include <Mouse.h>
#include <Keyboard.h>

// Number of pixels to move the mouse
#define PIXEL 3

// Method for shutting Windows down - available: winshutdown and winshutdown_alt
#define WINSHUTDOWN winshutdown

// Which input to use for the 2 Keys - 1 Locks the Screen, 2 Shuts Windows down, 3 disabled/enables the jiggler and changes the blink-frequency
#define KEY1 9
#define KEY2 11
#define KEY3 10

#define KEY_RELEASE 255

// Size of the fifo. lower would be enough, too but ram wants to be used...
#define FIFOSIZE 256

struct keybuffer {
  unsigned char c;
  int d;
  unsigned char s;
};

keybuffer fifo[FIFOSIZE+1];
int fifofirst=0;
int fifolast=0;

//const int ledpin =  LED_BUILTIN_RX;
const int ledpin =  LED_BUILTIN;

// LED timings in ms
const long led1 = 100;
const long led2 = 700;

// Key wait time in ms to avoid double pressing
const long key_1_rwait = 1000;
const long key_2_rwait = 10000;
const long key_3_rwait = 1000;

unsigned long currentms, lastms = 0;
int x, y, mx, my;
unsigned long jigglerwait;
unsigned long bufferwait;
unsigned long ledwait;
unsigned long key_1_wait;
unsigned long key_2_wait;
unsigned long key_3_wait;
int jigglerstate;
int ledstate;
int jigglerdisabled = 0;

// Initialize some variables - used more than once so predefining above is not sufficient
void initvars() {
    jigglerwait = 0;
    bufferwait = 0;
    ledwait = 0;
    jigglerstate = 0;
    ledstate = 0;
    key_1_wait = 0;
    key_2_wait = 0;
    key_3_wait = 0;
}

void setup() {
  pinMode(ledpin, OUTPUT);
  pinMode(KEY1, INPUT_PULLUP);
  pinMode(KEY2, INPUT_PULLUP);
  pinMode(KEY3, INPUT_PULLUP);

  // seed the RNG with some garbish vom AD0
  randomSeed(analogRead(0));
  Mouse.begin();
  Keyboard.begin();
//  Serial.begin(9600);
  initvars();
}

// Very very simple fifo without buffer-full-check
// Put one character plus wait-time after typing into fifo
void putfifo(char c, int d) {
  fifo[fifolast].c = c;
  fifo[fifolast].s = 0;
  fifo[fifolast].d = d;
  if (fifolast == FIFOSIZE) {
    fifolast = 0;
  } else {
    fifolast++;
  }
}

// put special key into queue - 255/KEY_RELEASE means: release all keys
void putfifosk(char s, int d) {
  fifo[fifolast].c = 0;
  fifo[fifolast].s = s;
  fifo[fifolast].d = d;
  if (fifolast == FIFOSIZE) {
    fifolast = 0;
  } else {
    fifolast++;
  }
}

// Add String c to fifo with delay of d ms after each character
void putfifos(char *c, int d) {
  int len,i=0;
  len=strlen(c);
  if (len >= FIFOSIZE) {
    len=FIFOSIZE-1;
  }
  for (i=0;i<len;i++) {
    putfifo(c[i],d);
  }
}

// Get one fifo entry
keybuffer getfifo() {
  int thi;
  if (! (fifofirst == fifolast)) {
    thi=fifofirst;
    if (fifofirst == FIFOSIZE) {
      fifofirst = 0;
    } else {
      fifofirst++;
    }
    return fifo[thi];;  
  }
}

// Returns state of fifo (true if empty)
int fifoempty() {
  return (fifofirst == fifolast);
}

// Writes contents of fifo to USB Keyboard one key by each other and wait the specified time after each character
// A character with value 0 means: no key but just wait or maybe a special key (if this value != 0)
void writebuffer() {
  keybuffer k;
  if (bufferwait < currentms ) {
    if ( ! fifoempty() ) {
      k=getfifo();
      if (k.c) {
        Keyboard.write(k.c);
      } else if (k.s) {
        if (k.s != KEY_RELEASE) {
          Keyboard.press(k.s);
        } else {
          Keyboard.releaseAll();
        }
      }
      bufferwait = currentms + k.d;
    }
  }
}

// Lock Windows Screen
void winlock() {
    putfifosk(KEY_LEFT_GUI,10);
    putfifo('l',10);
    putfifosk(KEY_RELEASE,50);
}

// Shut down Windows
void winshutdown() { 
    putfifosk(KEY_LEFT_GUI,10);
    putfifo('r',10);
    putfifosk(KEY_RELEASE,500);
    putfifos((char *) "cmd",50);
    putfifo(KEY_RETURN,150);
    putfifo(0,500);
    putfifos((char *) "shutdown /s /t 10 /f",50);
    putfifo(KEY_RETURN,150);
    putfifo(KEY_RETURN,50);
    putfifos((char *) "exit",50);
    putfifo(0,500);
    putfifo(KEY_RETURN,150);
}

void winshutdown_alt() { 
    putfifosk(KEY_LEFT_ALT,10);
    putfifosk(KEY_LEFT_CTRL,10);
    putfifosk(KEY_DELETE,10);
    putfifosk(KEY_RELEASE,500);
    putfifosk(KEY_LEFT_SHIFT,10);
    putfifosk(KEY_TAB,10);
    putfifosk(KEY_RELEASE,500);
    putfifo(KEY_RETURN,100);
    putfifo(KEY_DOWN_ARROW,500);
    putfifo(KEY_DOWN_ARROW,500);  
    putfifo(0,1000);
    putfifo(KEY_RETURN,150);
}

// Heartbeat the given led
void ledblink() {
  int value;
  if (currentms >= ledwait) {
    switch (ledstate) {
      case 0:
        value=HIGH;
        ledwait=currentms+led1+(jigglerdisabled*led1*2);
        ledstate++;
        break;
      case 1:
        value=LOW;
        ledwait=currentms+led1+(jigglerdisabled*led1*2);
        ledstate++;
        break;
      case 2:
        value=HIGH;
        ledwait=currentms+led1+(jigglerdisabled*led1*2);
        ledstate++;
        break;
      default:
        value=LOW;
        ledwait=currentms+led2+(jigglerdisabled*led2*2);
        ledstate=0;
        break;
    }
    digitalWrite(ledpin, value);
  }
}

// Jiggles the mouse - with random width (configure the maximum PIXEL above) and random time between 2 jiggles
void jiggle() {
  if (currentms >= jigglerwait) {
    switch (jigglerstate) {
      case 0:
        if (jigglerdisabled == 0) { 
          //random-generatre move, initial move
          x = random(-PIXEL, PIXEL);
          y = random(-PIXEL, PIXEL);
          mx = 0 - x;
          my = 0 - y;
          Mouse.move(x, y, 0);
          jigglerwait = currentms + random(50, 150);
          jigglerstate = 1;
        } else {
          jigglerwait = currentms + 500;
        }
        break;
      case 1:
        // move backwards;
        Mouse.move(mx, my, 0);
        jigglerwait = currentms + random(50, 150);
        jigglerstate = 2;
        break;
      case 2:
        // move backwards again;
        Mouse.move(mx, my, 0);
        jigglerwait = currentms + random(50, 150);
        jigglerstate = 3;
        break;
      case 3:
        // move to original position;
        Mouse.move(x, y, 0);
        jigglerwait = currentms + random(5000, 15000);
        jigglerstate = 0;
        break;
      default:
        jigglerstate = 0;
        break;
    }
  }
}

// Check key1 and do winlock() if pressed
void key1() {
  if (digitalRead(KEY1) != HIGH) {
    if (currentms >= key_1_wait ) {
      key_1_wait = currentms + key_1_rwait;
      if (key_2_wait < key_1_wait) {
        key_2_wait = key_1_rwait;
      }
      winlock();
    }
  }
}

// Check key2 and do winshutdown if pressed
void key2() {
  if (digitalRead(KEY2) != HIGH) {
    if (currentms >= key_2_wait ) {
      key_2_wait = currentms + key_2_rwait;
      WINSHUTDOWN();
    }
  }
}

// Check key3 and change jigglerstatus if pressed
void key3() {
  if (digitalRead(KEY3) != HIGH) {
    if (currentms >= key_3_wait ) {
      key_3_wait = currentms + key_3_rwait;
      if (jigglerdisabled) {
        jigglerdisabled = 0;
      } else {
        jigglerdisabled = 1;
      }
    }
  }
}

// Main loop - its some sort of a cooperative multitasking
void loop() {
  currentms = millis();
  if (currentms < lastms) {
    initvars();
  }
  key1();
  key2();
  key3();
  ledblink();
  jiggle();
  writebuffer();
  lastms = currentms;
  delay(5);
}
